"""
atomSmltr
================

a python package for atomic physics simulations.

Package structure
------------------

| ``atomsmltr.atoms``        : atomic species handling (physical properties, electronic transitions)
| ``atomsmltr.datapipe``     : manage data generated by the simulations (save, load..)
| ``atomsmltr.environment``  : define the atom environment (lasers, mag. fields, zones)
| ``atomsmltr.simulation``   : manage simulation configuration & run simulations
| ``atomsmltr.utils``        : misc tools
| ``atomsmltr.examples``     : contain pre-made example configurations

"""

__version__ = "0.0.1"
__author__ = "Alexandre Dareau"
__all__ = [
    "atoms",
    "datapipe",
    "lasers",
    "zones",
    "fields",
    "simulation",
    "utils",
    "examples",
]

from atomsmltr import atoms
from atomsmltr import datapipe
from atomsmltr.environment import lasers, zones, fields
from atomsmltr import simulation
from atomsmltr import utils
from atomsmltr import examples
