"""The ``atomsmltr.examples`` subpackage provides a collection of example configurations
for illustration & testing purposes

Examples
---------

Import the configuration from Feng et al 2024 Quantum Sci. Technol. 9 025017

.. code-block:: python

    from atomsmltr.examples.feng2024 import config
    from atomsmltr.examples.feng2024 import description
    print(description)

"""
